{*******************************************************}
{File:      NCOciUpdateSQL.PAS                          }
{Revision:  0.02.02 / 23.04.2000                        }
{Comment:   NC OCI8 VCL: editing capabilities           }
{Copyright: (c) 2000, Dmitry Arefiev                    }
{Author:    Dmitry Arefiev, dmitrya@inthink.com         }
{*******************************************************}
{$I NCOciDef.inc}

unit NCOciUpdateSQL;

interface

uses Windows, SysUtils, Graphics, Classes, Controls, Db, NCOci, NCOciWrapper,
     NCOciDB, NCOciParams;

type
    TOCIUpdateSQL = class;

    TOCILockMode = (lmPesimistic, lmOptimistic, lmNone);
    TOCILockPoint = (lpImmediate, lpDefered);
    TOCIUpdateMode = (upWhereAll, upWhereChanged, upWhereKeyOnly);

    TOCIUpdateSQLEvent = procedure (ASender: TOCIUpdateSQL; AKind: TOCISQLKind;
        var ADefault: Boolean) of object;

    TOCIUpdateSQL = class(TOCIUpdateObject)
    private
        FQueries: array[TOCISQLKind] of TOCIQuery;
        FSQLText: array[TOCISQLKind] of TStrings;
        FSQLParams: array[TOCISQLKind] of TOCIParams;
        FUserDefinedKinds: TOCISQLKinds;
        FSelfUpdating: Boolean;
        FTableName: String;
        FUpdateChangedFields: Boolean;
        FAutoBuildKinds: TOCISQLKinds;
        FLockMode: TOCILockMode;
        FLockPoint: TOCILockPoint;
        FUpdateMode: TOCIUpdateMode;
        FOnSQLGenerate: TOCIUpdateSQLEvent;
        FOnApply: TOCIUpdateSQLEvent;
        FOnDisconnect: TOCIUpdateSQLEvent;
        function GetQuery(ASQLKind: TOCISQLKind): TOCIQuery;
        function GetSQLIndex(Index: Integer): TStrings;
        procedure SetSQL(ASQLKind: TOCISQLKind; Value: TStrings);
        procedure SetSQLIndex(Index: Integer; Value: TStrings);
        function IsSQLStored(Index: Integer): Boolean;
        function GetSQLParams(ASQLKind: TOCISQLKind): TOciParams;
        procedure SetSQLParams(ASQLKind: TOCISQLKind; const Value: TOciParams);
        function GetActualTableName: String;
        function IsOraBlob(AField: TField): Boolean;
        function IsROWID(AField: TField): Boolean;
        procedure SetTableName(const Value: String);
        function Par2FieldName(var AName: String): Integer;
        procedure SetUpdateMode(const Value: TOCIUpdateMode);
        procedure InternalApply(ASQLKind: TOCISQLKind);
        procedure SetLockMode(const Value: TOCILockMode);
        procedure SetLockPoint(const Value: TOCILockPoint);
    protected
        function BuildFieldParam(AField: TField; ASQLKind: TOCISQLKind;
            AParType: TOCIVarType; ANewValue: Boolean): String;
        function BuildFieldSQL(AField: TField): String;
        procedure BuildWherePhrase(ASQLKind: TOCISQLKind);
        procedure BuildReturningPhrase(ASQLKind: TOCISQLKind);
        function GetSQL(ASQLKind: TOCISQLKind): TStrings;
        procedure SQLChanged(Sender: TObject);
        function GenerateLock: Boolean; virtual;
        function GenerateUpdate: Boolean; virtual;
        function GenerateInsert: Boolean; virtual;
        function GenerateDelete: Boolean; virtual;
        function GenerateRefresh: Boolean; virtual;
        function FieldStorable(AField: TField): Boolean; virtual;
        function FieldUpdatable(AField: TField): Boolean; virtual;
        function FieldLocatable(AField: TField): Boolean; virtual;
        function FieldReqRefresh(AField: TField): Boolean; virtual;
        function FieldChanged(AField: TField): Boolean; virtual;
        function FieldInKey(AField: TField): Boolean; virtual;
        function GetSupported(ASQLKind: TOCISQLKind): Boolean; override;
    public
        constructor Create(AOwner: TComponent); override;
        destructor Destroy; override;
        procedure Apply(ASQLKind: TOCISQLKind); override;
        procedure DisconnectAll; override;
        procedure ExecSQL(ASQLKind: TOCISQLKind);
        procedure SetParams(ASQLKind: TOCISQLKind);
        procedure RefreshData(ASQLKind: TOCISQLKind);
        procedure Disconnect(ASQLKind: TOCISQLKind);
        property Query[ASQLKind: TOCISQLKind]: TOCIQuery read GetQuery;
        property SQL[ASQLKind: TOCISQLKind]: TStrings read GetSQL write SetSQL;
        property SQLParams[ASQLKind: TOCISQLKind]: TOciParams read GetSQLParams
            write SetSQLParams;
        property UserDefinedKinds: TOCISQLKinds read FUserDefinedKinds;
        property ActualTableName: String read GetActualTableName;
    published
        property UpdateChangedFields: Boolean read FUpdateChangedFields
            write FUpdateChangedFields default True;
        property AutoBuildKinds: TOCISQLKinds read FAutoBuildKinds write FAutoBuildKinds
            default [skLock, skUnLock, skUpdate, skInsert, skDelete, skRefresh];
        property LockMode: TOCILockMode read FLockMode write SetLockMode default lmOptimistic;
        property LockPoint: TOCILockPoint read FLockPoint write SetLockPoint default lpDefered;
        property UpdateMode: TOCIUpdateMode read FUpdateMode write SetUpdateMode default upWhereKeyOnly;
        property TableName: String read FTableName write SetTableName;
        property SQLLock: TStrings index 0 read GetSQLIndex write SetSQLIndex stored IsSQLStored;
        property SQLUnLock: TStrings index 1 read GetSQLIndex write SetSQLIndex stored IsSQLStored;
        property SQLUpdate: TStrings index 2 read GetSQLIndex write SetSQLIndex stored IsSQLStored;
        property SQLInsert: TStrings index 3 read GetSQLIndex write SetSQLIndex stored IsSQLStored;
        property SQLDelete: TStrings index 4 read GetSQLIndex write SetSQLIndex stored IsSQLStored;
        property SQLRefresh: TStrings index 5 read GetSQLIndex write SetSQLIndex stored IsSQLStored;
        property OnSQLGenerate: TOCIUpdateSQLEvent read FOnSQLGenerate write FOnSQLGenerate;
        property OnApply: TOCIUpdateSQLEvent read FOnApply write FOnApply;
        property OnDisconnect: TOCIUpdateSQLEvent read FOnDisconnect write FOnDisconnect; 
    end;

implementation

Uses NCOciUtil, NCOciMsg;

const
    SSQLKinds: array[TOCISQLKind] of String =
        ('Lock', 'Unlock', 'Update', 'Insert', 'Delete', 'Refresh',
         'LockImmediate', 'UnLockImmediate');

constructor TOCIUpdateSQL.Create(AOwner: TComponent);
var
    SQLKind: TOCISQLKind;
begin
    inherited Create(AOwner);
    FUpdateChangedFields := True;
    FAutoBuildKinds := [skLock, skUnLock, skUpdate, skInsert, skDelete, skRefresh];
    FLockMode := lmOptimistic;
    FLockPoint := lpDefered;
    FUpdateMode := upWhereKeyOnly;
    for SQLKind := Low(TOCISQLKind) to High(TOCISQLKind) do begin
        FSQLText[SQLKind] := TStringList.Create;
        TStringList(FSQLText[SQLKind]).OnChange := SQLChanged;
        FSQLParams[SQLKind] := TOciParams.Create;
    end;
end;

destructor TOCIUpdateSQL.Destroy;
var
    SQLKind: TOCISQLKind;
begin
    for SQLKind := Low(TOCISQLKind) to High(TOCISQLKind) do begin
        FSQLText[SQLKind].Free;
        FSQLParams[SQLKind].Free;
    end;
    inherited Destroy;
end;

function TOCIUpdateSQL.IsSQLStored(Index: Integer): Boolean;
begin
    Result := TOCISQLKind(Index) in FUserDefinedKinds;
end;

function TOCIUpdateSQL.GetQuery(ASQLKind: TOCISQLKind): TOCIQuery;
begin
    if (FQueries[ASQLKind] = nil) and (DataSet <> nil) then begin
        FQueries[ASQLKind] := TOCIQuery.Create(Self);
        with FQueries[ASQLKind] do begin
            Disconnectable := True;
            DatabaseName := DataSet.DatabaseName;
            TransactionManager := DataSet.TransactionManager;
            FetchParams := DataSet.FetchParams;
            FetchParams.FetchAll := True;
            DataFormat := DataSet.DataFormat;
            SQL.BeginUpdate;
            try
                SQL.Assign(Self.SQL[ASQLKind]);
                if DataSet is TOCICustomQuery then
                    Macros.AssignValues(TOCIQuery(DataSet).Macros);
            finally
                SQL.EndUpdate;
            end;
            Params.AssignValues(Self.SQLParams[ASQLKind]);
            if DataSet is TOCICustomQuery then
                Params.AssignValues(TOCICustomQuery(DataSet).Params);
        end;
    end;
    Result := FQueries[ASQLKind];
end;

function TOCIUpdateSQL.GetSQL(ASQLKind: TOCISQLKind): TStrings;
var
    canGenerate, genDef, workExist: Boolean;
begin
    if ([csDesigning, csReading] * ComponentState = []) and
       (FSQLText[ASQLKind].Count = 0) and
       (ASQLKind in AutoBuildKinds) then begin
        FSelfUpdating := True;
        try
            FSQLText[ASQLKind].Clear;
            FSQLParams[ASQLKind].Clear;
            genDef := True;
            workExist := True;
            if Assigned(OnSQLGenerate) then
                OnSQLGenerate(Self, ASQLKind, genDef);
            if genDef then begin
                canGenerate := False;
                workExist := False;
                if DataSet is TOCICustomQuery then
                    with TOCICustomQuery(DataSet) do begin
                        if not Prepared then
                            Prepare;
                        canGenerate := (StatementType = stSelect);
                    end;
                if canGenerate then
                    case ASQLKind of
                        skLock: workExist := GenerateLock;
                        skUpdate: workExist := GenerateUpdate;
                        skInsert: workExist := GenerateInsert;
                        skDelete: workExist := GenerateDelete;
                        skRefresh: workExist := GenerateRefresh;
                    end;
            end;
            if (FSQLText[ASQLKind].Count = 0) and workExist then
                OCIDBErrorFmt(msgOCICantGenQuery, [SSQLKinds[ASQLKind]]);
        finally
            FSelfUpdating := False;
        end;
    end;
    Result := FSQLText[ASQLKind];
end;

procedure TOCIUpdateSQL.SetSQL(ASQLKind: TOCISQLKind; Value: TStrings);
begin
    FSQLText[ASQLKind].Assign(Value);
end;

function TOCIUpdateSQL.GetSQLIndex(Index: Integer): TStrings;
begin
    Result := GetSQL(TOCISQLKind(Index));
end;

procedure TOCIUpdateSQL.SetSQLIndex(Index: Integer; Value: TStrings);
begin
    SetSQL(TOCISQLKind(Index), Value);
end;

function TOCIUpdateSQL.GetSQLParams(ASQLKind: TOCISQLKind): TOciParams;
begin
    Result := FSQLParams[ASQLKind];
end;

procedure TOCIUpdateSQL.SetSQLParams(ASQLKind: TOCISQLKind;
  const Value: TOciParams);
begin
    FSQLParams[ASQLKind].Assign(Value);
end;

procedure TOCIUpdateSQL.SQLChanged(Sender: TObject);
var
    SQLKind: TOCISQLKind;
begin
    for SQLKind := Low(TOCISQLKind) to High(TOCISQLKind) do
        if Sender = FSQLText[SQLKind] then begin
            if not FSelfUpdating then begin
                Include(FUserDefinedKinds, SQLKind);
                Disconnect(SQLKind);
            end;
            Break;
        end;
end;

function TOCIUpdateSQL.GetActualTableName: String;
begin
    if FTableName <> '' then
        Result := FTableName
    else if DataSet <> nil then
        Result := DataSet.TableName
    else
        Result := '';
end;

procedure TOCIUpdateSQL.SetTableName(const Value: String);
begin
    if ActualTableName <> Value then begin
        FTableName := Value;
        DisconnectAll;
    end;
end;

procedure TOCIUpdateSQL.SetUpdateMode(const Value: TOCIUpdateMode);
begin
    if UpdateMode <> Value then begin
        FUpdateMode := Value;
        Disconnect(skUpdate);
    end;
end;

procedure TOCIUpdateSQL.SetLockMode(const Value: TOCILockMode);
begin
    CheckNoUpdatesPending;
    FLockMode := Value;
end;

procedure TOCIUpdateSQL.SetLockPoint(const Value: TOCILockPoint);
begin
    CheckNoUpdatesPending;
    FLockPoint := Value;
end;

function TOCIUpdateSQL.GetSupported(ASQLKind: TOCISQLKind): Boolean;
begin
    if Assigned(OnSQLGenerate) then
        Result := True
    else if DataSet is TOCICustomQuery then
        with TOCICustomQuery(DataSet) do begin
            if not Prepared then
                Prepare;
            Result := (StatementType = stSelect);
        end
    else
        Result := False;
    Result := Result and not (
       (DataSet = nil) or
       (DataSet.Database = nil) or
       (ASQLKind in [skLockImmediate, skUnLockImmediate]) and
            ((LockMode = lmNone) or (LockPoint = lpDefered)) or
       (ASQLKind in [skLock, skUnLock]) and
            ((LockMode = lmNone) or (LockPoint = lpImmediate) or
             (LockMode = lmOptimistic) and (LockPoint = lpDefered))
    );
end;

function TOCIUpdateSQL.FieldChanged(AField: TField): Boolean;
begin
    Result := not VarNearEqual(AField.NewValue, AField.OldValue);
end;

function TOCIUpdateSQL.FieldStorable(AField: TField): Boolean;
begin
    Result := (AField.FieldKind = fkData) and (AField.Origin <> '');
end;

function TOCIUpdateSQL.FieldUpdatable(AField: TField): Boolean;
begin
    Result := FieldStorable(AField) and not AField.ReadOnly and
        (pfInUpdate in AField.ProviderFlags);
end;

function TOCIUpdateSQL.FieldLocatable(AField: TField): Boolean;
begin
    Result := FieldStorable(AField) and not (DataSet.ODataType[AField.FieldNo] in
        [otLong, otLongRaw, otCursor, otCLOB, otBLOB, otCFile, otBFile]) and
        (AField.ProviderFlags * [pfInKey, pfInWhere] <> []);
end;

function TOCIUpdateSQL.FieldReqRefresh(AField: TField): Boolean;
begin
{$IFDEF OCI_D5}
    Result := (AField.AutoGenerateValue = arAutoInc) or
              (AField.AutoGenerateValue = arDefault) and AField.IsNull;
{$ELSE}
    Result := False;
{$ENDIF}
end;

function TOCIUpdateSQL.IsOraBlob(AField: TField): Boolean;
begin
    Result := (DataSet.ODataType[AField.FieldNo] in
        [otBLOB, otCLOB, otBFILE, otCFILE]);
end;

function TOCIUpdateSQL.IsROWID(AField: TField): Boolean;
begin
    Result := (DataSet.ODataType[AField.FieldNo] in
        [otROWID]);
end;

function TOCIUpdateSQL.FieldInKey(AField: TField): Boolean;
begin
    Result := (pfInKey in AField.ProviderFlags);
end;

function TOCIUpdateSQL.BuildFieldParam(AField: TField; ASQLKind: TOCISQLKind;
    AParType: TOCIVarType; ANewValue: Boolean): String;
const
    valAge: array[Boolean] of String = ('OLD_', 'NEW_');
var
    par: TOCIParam;
    oParName: String;
begin
    Result := ':"' + valAge[ANewValue] + AField.FieldName + '"';
    oParName := ':' + valAge[ANewValue] + AField.FieldName;
    par := FSQLParams[ASQLKind].FindParam(oParName);
    if par <> nil then begin
        if par.OParamType <> AParType then
            par.OParamType := odInOut;
    end
    else
        FSQLParams[ASQLKind].CreateOParam(DataSet.ODataType[AField.FieldNo],
            DataSet.ODataSize[AField.FieldNo], oParName, AParType);
end;

function TOCIUpdateSQL.BuildFieldSQL(AField: TField): String;
begin
    Result := '"' + AField.Origin + '"';
end;

procedure TOCIUpdateSQL.BuildWherePhrase(ASQLKind: TOCISQLKind);
var
    s: String;
    i: Integer;
begin
    s := '';
    for i := 0 to DataSet.FieldCount - 1 do
        if FieldLocatable(DataSet.Fields[i]) and
           ((UpdateMode = upWhereAll) or
            (UpdateMode = upWhereChanged) and FieldChanged(DataSet.Fields[i]) or
            (UpdateMode = upWhereKeyOnly) and FieldInKey(DataSet.Fields[i])) then begin
            if s <> '' then
                s := s + ' and ';
            if VarIsNull(DataSet.Fields[i].OldValue) then
                s := s + BuildFieldSQL(DataSet.Fields[i]) + ' is null'
            else
                s := s + BuildFieldSQL(DataSet.Fields[i]) + ' = ' +
                         BuildFieldParam(DataSet.Fields[i], ASQLKind, odIn, False);
        end;
    if s = '' then
        OCIDBError(msgOCIKeyFieldsEmpty);
    FSQLText[ASQLKind].Add('where ' + s);
end;

procedure TOCIUpdateSQL.BuildReturningPhrase(ASQLKind: TOCISQLKind);
var
    s1, s2: String;
    i: Integer;
begin
    s1 := '';
    s2 := '';
    for i := 0 to DataSet.FieldCount - 1 do
        if FieldStorable(DataSet.Fields[i]) and
           (FieldReqRefresh(DataSet.Fields[i]) or
            (ASQLKind = skInsert) and (IsOraBlob(DataSet.Fields[i]) or IsROWID(DataSet.Fields[i]))) then begin
            if s1 <> '' then begin
                s1 := s1 + ', ';
                s2 := s2 + ', ';
            end;
            s1 := s1 + BuildFieldSQL(DataSet.Fields[i]);
            s2 := s2 + BuildFieldParam(DataSet.Fields[i], ASQLKind, odRet, True);
        end;
    if s1 <> '' then
        FSQLText[ASQLKind].Add('returning ' + s1 + ' into ' + s2);
end;

function TOCIUpdateSQL.GenerateLock: Boolean;
begin
    FSQLText[skLock].Add('select * from ' + ActualTableName);
    BuildWherePhrase(skLock);
    if LockMode = lmPesimistic then
        FSQLText[skLock].Add('for update nowait');
    Result := True;
end;

function TOCIUpdateSQL.GenerateUpdate: Boolean;
var
    i: Integer;
    s: String;
    fldChng: Boolean;
begin
    s := '';
    Result := False;
    for i := 0 to DataSet.FieldCount - 1 do begin
        fldChng := FieldChanged(DataSet.Fields[i]);
        Result := Result or fldChng;
        if FieldUpdatable(DataSet.Fields[i]) and not IsOraBlob(DataSet.Fields[i]) and
           (not UpdateChangedFields or fldChng) then begin
            if s <> '' then
                s := s + ', ';
            s := s + BuildFieldSQL(DataSet.Fields[i]) + ' = ' +
                     BuildFieldParam(DataSet.Fields[i], skUpdate, odIn, True);
        end;
    end;
    if s <> '' then begin
        FSQLText[skUpdate].Add('update ' + ActualTableName);
        FSQLText[skUpdate].Add('set ' + s);
        BuildWherePhrase(skUpdate);
        BuildReturningPhrase(skUpdate);
    end;
end;

function TOCIUpdateSQL.GenerateInsert: Boolean;
var
    i: Integer;
    s1, s2, s3: String;
    fldChng: Boolean;
begin
    s1 := '';
    s2 := '';
    Result := False;
    for i := 0 to DataSet.FieldCount - 1 do begin
        fldChng := IsOraBlob(DataSet.Fields[i]) or not UpdateChangedFields or
            not DataSet.Fields[i].IsNull;
        Result := Result or fldChng;
        if FieldUpdatable(DataSet.Fields[i]) then begin
            if (s3 = '') and not DataSet.Fields[i].Required then
                s3 := BuildFieldSQL(DataSet.Fields[i]);
            if fldChng then begin
                if s1 <> '' then begin
                    s1 := s1 + ', ';
                    s2 := s2 + ', ';
                end;
                s1 := s1 + BuildFieldSQL(DataSet.Fields[i]);
                case DataSet.ODataType[DataSet.Fields[i].FieldNo] of
                    otBLOB, otBFILE: s2 := s2 + 'EMPTY_BLOB()';
                    otCLOB, otCFILE: s2 := s2 + 'EMPTY_CLOB()';
                    else s2 := s2 + BuildFieldParam(DataSet.Fields[i], skInsert, odIn, True);
                end;
            end;
        end;
    end;
    FSQLText[skInsert].Add('insert into ' + ActualTableName);
    if s1 = '' then begin
        s1 := s3;
        s2 := 'NULL';
    end;
    if s1 <> '' then begin
        FSQLText[skInsert].Add('(' + s1 + ')');
        FSQLText[skInsert].Add('values (' + s2 + ')');
        BuildReturningPhrase(skInsert);
    end;
end;

function TOCIUpdateSQL.GenerateDelete: Boolean;
begin
    FSQLText[skDelete].Add('delete from ' + ActualTableName);
    BuildWherePhrase(skDelete);
    Result := True;
end;

function TOCIUpdateSQL.GenerateRefresh: Boolean;
begin
    FSQLText[skRefresh].Add('select * from ' + ActualTableName);
    BuildWherePhrase(skRefresh);
    Result := True;
end;

procedure TOCIUpdateSQL.ExecSQL(ASQLKind: TOCISQLKind);
begin
    if DataSet <> nil then
        with Query[ASQLKind] do
        try
            if SQL.Count = 0 then
                Exit;
            if not Prepared then
                Prepare;
            if (ASQLKind = skLock) and (LockMode = lmPesimistic) then
                Database.StartTransaction;
            if StatementType = stSelect then begin
                Close;
                Open;
            end
            else
                ExecSQL;
            if (StatementType in [stUpdate, stDelete, stInsert]) and (RowsAffected <> 1) or
               (StatementType in [stSelect]) and (RecordCount <> 1) then
                OCIDBError(msgOCIRecordDeleted);
            if (ASQLKind = skLock) and (StatementType in [stSelect]) and
               not DataSet.RecordsEqual(Query[ASQLKind], dsOldValue, dsCurValue) then begin
                Apply(skUnLock);
                OCIDBError(msgOCIRecordChanged);
            end;
        except on E: EOCINativeError do
            // resource busy and acquire with NOWAIT specified
            if (E.Errors[0].ErrorCode = 54) then
                OCIDBError(msgOCIRecordLocked)
            else
                raise;
        end;
end;

procedure TOCIUpdateSQL.DisconnectAll;
var
    SQLKind: TOCISQLKind;
begin
    for SQLKind := Low(TOCISQLKind) to High(TOCISQLKind) do
        Disconnect(SQLKind);
end;

procedure TOCIUpdateSQL.Disconnect(ASQLKind: TOCISQLKind);
var
    doDef: Boolean;
begin
    doDef := True;
    if Assigned(OnDisconnect) then
        OnDisconnect(Self, ASQLKind, doDef);
    if doDef then begin
        if not (ASQLKind in FUserDefinedKinds) then begin
            FSelfUpdating := True;
            try
                FSQLText[ASQLKind].Clear;
            finally
                FSelfUpdating := False;
            end;
        end;
        if FQueries[ASQLKind] <> nil then begin
            FQueries[ASQLKind].Free;
            FQueries[ASQLKind] := nil;
        end;
    end;
end;

function TOCIUpdateSQL.Par2FieldName(var AName: String): Integer;
begin
    if CompareText(Copy(AName, 1, 4), 'OLD_') = 0 then begin
        Result := -1;
        System.Delete(AName, 1, 4);
    end
    else if CompareText(Copy(AName, 1, 4), 'NEW_') = 0 then begin
        Result := 1;
        System.Delete(AName, 1, 4);
    end
    else
        Result := 1;
end;

procedure TOCIUpdateSQL.SetParams(ASQLKind: TOCISQLKind);
var
    I: Integer;
    Old: Boolean;
    Param: TOCIParam;
    parName: string;
    Field: TField;
    Value: Variant;
begin
    if DataSet <> nil then
        with Query[ASQLKind] do
            for I := 0 to Params.Count - 1 do begin
                Param := Params[I];
                if Param.OParamType in [odUnknown, odIn, odInOut] then begin
                    if Param.OParamType = odUnknown then
                        Param.OParamType := odIn;
                    parName := Param.Name;
                    Old := Par2FieldName(parName) = -1;
                    Field := DataSet.FindField(parName);
                    if Field <> nil then
                        if Old then
                            Param.AssignFieldValue(Field, Field.OldValue)
                        else begin
                            Value := Field.NewValue;
                            if VarIsEmpty(Value) then
                                Value := Field.OldValue;
                            Param.AssignFieldValue(Field, Value);
                        end;
                end;
            end;
end;

type
    __TOCIDataSet = class(TOCIDataSet);

procedure TOCIUpdateSQL.RefreshData(ASQLKind: TOCISQLKind);
var
    prevState: TDataSetState;
    i: Integer;
    new: Boolean;
    param: TOCIParam;
    parName: string;
    field: TField;
begin
    if DataSet <> nil then
        with Query[ASQLKind] do begin
            if SQL.Count = 0 then
                Exit;
            prevState := __TOCIDataSet(DataSet).SetTempState(dsNewValue);
            try
                if StatementType = stSelect then
                    for i := 0 to FieldCount - 1 do begin
                        field := DataSet.FindField(Fields[i].FieldName);
                        if field <> nil then
                            field.Assign(Fields[i]);
                    end
                else if StatementType in [stDeclare, stBegin, stInsert, stDelete, stUpdate] then
                    for i := 0 to Params.Count - 1 do begin
                        param := Params[I];
                        parName := param.Name;
                        new := Par2FieldName(parName) = 1;
                        if (param.OParamType in [odOut, odInOut, odRet]) and new then begin
                            field := DataSet.FindField(parName);
                            if field <> nil then
                                field.Assign(param);
                        end;
                    end
                else
                    OCIDBError(msgOCICantUseSQL4Refresh);
            finally
                __TOCIDataSet(DataSet).RestoreState(prevState);
            end;
        end;
end;

procedure TOCIUpdateSQL.InternalApply(ASQLKind: TOCISQLKind);
var
    appDef: Boolean;
begin
    appDef := True;
    if Assigned(OnApply) then
        OnApply(Self, ASQLKind, appDef);
    if appDef then
        if ASQLKind = skLockImmediate then
            InternalApply(skLock)
        else if ASQLKind = skUnLockImmediate then
            InternalApply(skUnLock)
        else if (ASQLKind = skUnLock) and (skUnLock in AutoBuildKinds) then begin
            if LockMode = lmPesimistic then
                DataSet.Database.Rollback;
        end
        else begin
            if UpdateChangedFields and (ASQLKind in [skUpdate, skInsert]) or
               (UpdateMode = upWhereChanged) and (ASQLKind = skUpdate) or
               (UpdateMode <> upWhereKeyOnly) and
                    (ASQLKind in [skLock, skUpdate, skDelete, skRefresh]) then
                Disconnect(ASQLKind);
            GetSQL(ASQLKind);
            GetQuery(ASQLKind);
            SetParams(ASQLKind);
            ExecSQL(ASQLKind);
            if not (ASQLKind in [skLock, skUnlock, skLockImmediate, skUnLockImmediate]) then
                RefreshData(ASQLKind);
        end;
end;

procedure TOCIUpdateSQL.Apply(ASQLKind: TOCISQLKind);
begin
    if not Supported[ASQLKind] then
        Exit;
    InternalApply(ASQLKind);
end;

end.
